@@ -6,7 +6,7 @@ from typing import Callable, Dict, Optional, Tuple, List, Mapping
 import numpy as np
 import torch
 from catalyst.contrib.tools.tensorboard import SummaryWriter
-from catalyst.core import Callback, CallbackOrder, State, IRunner
+from catalyst.core import Callback, CallbackOrder, State, IRunner, BatchMetricCallback
 from torch import nn
 from torch.nn import DataParallel
 from torch.utils.data import Dataset, SequentialSampler
@@ -64,7 +64,7 @@ def img(writer: SummaryWriter, idx: int, sample: Tuple[Dict, Dict], prefix: str,
     :param key: The key in X that is being published.
     """
     X, y = sample
-    writer.add_image(f'{prefix}_{task_name}_images', X[key])
+    writer.add_image(f'{prefix}_{task_name}_images', X[key], global_step=idx)
 
 
 def text(writer: SummaryWriter, idx: int, sample: Tuple, prefix: str, task_name: str, key: str):
@@ -84,7 +84,7 @@ def text(writer: SummaryWriter, idx: int, sample: Tuple, prefix: str, task_name:
     :param key: The key in X that is being published.
     """
     X, y = sample
-    writer.add_text(f'{prefix}_{task_name}_text', X[key])
+    writer.add_text(f'{prefix}_{task_name}_text', X[key], global_step=idx)
 
 
 def default_tensorboard_type_mapping():
@@ -203,6 +203,14 @@ class TensorboardConverters:
             logger.close()
 
 
+def should_skip_loader(state, loaders_to_skip):
+    if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
+        return True
+    if state.loader_name in loaders_to_skip:
+        return True
+    return False
+
+
 class InterpretationCallback(Callback):
     """
     This callback publishes best and worst images per task, according to the configuration supplied via the constructor.
@@ -240,9 +248,7 @@ class InterpretationCallback(Callback):
         return interpretation_dict
 
     def on_loader_start(self, state: State):
-        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
-            return
-        if state.loader_name in self.loaders_to_skip:
+        if should_skip_loader(state, self.loaders_to_skip):
             return
         self.interpretations[state.loader_name] = self._initialize_interpretations()
         self.loader_counts[state.loader_name] = 0
@@ -251,9 +257,7 @@ class InterpretationCallback(Callback):
             self.tensorboard_converters.initialize(state)
 
     def on_batch_end(self, state: State):
-        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
-            return
-        if state.loader_name in self.loaders_to_skip:
+        if should_skip_loader(state, self.loaders_to_skip):
             return
         outputs = state.output['logits']
         targets = state.input['targets']
@@ -271,9 +275,7 @@ class InterpretationCallback(Callback):
         self.loader_counts[state.loader_name] += bs
 
     def on_loader_end(self, state: State):
-        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
-            return
-        if state.loader_name in self.loaders_to_skip:
+        if should_skip_loader(state, self.loaders_to_skip):
             return
         self.interpretations[state.loader_name] = {
             key: np.concatenate(value, axis=0)
@@ -284,9 +286,7 @@ class InterpretationCallback(Callback):
             self.tensorboard_converters.publish(state, self.interpretations[state.loader_name])
 
     def on_stage_end(self, state: State):
-        if not isinstance(state.loaders[state.loader_name].sampler, SequentialSampler):
-            return
-        if state.loader_name in self.loaders_to_skip:
+        if should_skip_loader(state, self.loaders_to_skip):
             return
         if self.tensorboard_converters is not None:
             self.tensorboard_converters.close(state)
@@ -294,7 +294,7 @@ class InterpretationCallback(Callback):
 
 class DeviceReducingDataParallel(DataParallel):
 
-    def __init__(self, module: nn.Module, task_flow, callbacks):
+    def __init__(self, module: nn.Module, task_flow, infer_dict_callback):
         super().__init__(module)
         tasks_dict = task_flow.get_all_children()
         self.full_paths = []
@@ -302,14 +302,24 @@ class DeviceReducingDataParallel(DataParallel):
             if task.is_train_only():
                 continue
             self.full_paths.append(full_path)
+        criterion = task_flow.get_loss()
+        per_sample_criterion = task_flow.get_per_sample_loss(ctx=criterion.get_device_reduced_ctx())
+        leaf_losses = criterion.get_leaf_losses()
+        metrics = criterion.get_metrics()
         self._reducing_func = partial(reduce_on_device,
-                                      criterion=task_flow.get_loss(),
-                                      per_sample_criterion=task_flow.get_per_sample_loss(),
-                                      leaf_criterions=task_flow.get_loss().get_leaf_losses(),
-                                      metrics=task_flow.get_loss().get_metrics())
-        self.callbacks = callbacks
+                                      criterion=criterion,
+                                      per_sample_criterion=per_sample_criterion,
+                                      leaf_criterions=leaf_losses,
+                                      metrics=metrics)
+        self.infer_dict_callback = infer_dict_callback
+        self.ctx = criterion.ctx
+        self.r_device_metrics = False
+        self.r_leaf_losses = False
+        self.r_per_sample_losses = False
+        self.store_inference_results = self.infer_dict_callback is not None
 
     def gather(self, outputs, output_device):
+        self.ctx.clear()
         device_reduced_results = []
         dct = {
             'gt': {'_targets': {}}
@@ -318,36 +328,69 @@ class DeviceReducingDataParallel(DataParallel):
             dct[full_path] = []
             dct[f'precondition|{full_path}'] = []
 
+        ctx_reductions = {}
         for i in range(len(outputs)):
-            r = self._reducing_func(outputs=outputs[i], targets=outputs[i]['gt']['_targets'])
-            device_reduced_results.append(r)
-            for full_path in self.full_paths:
-                dct[full_path].append(outputs[i][full_path].detach().cpu().numpy())
-                precondition_path = f'precondition|{full_path}'
-                dct[precondition_path].append(outputs[i][precondition_path].detach().cpu().numpy())
-                np_targets = outputs[i]['gt']['_targets'][full_path].detach().cpu().numpy()
-                if full_path not in dct['gt']['_targets']:
-                    dct['gt']['_targets'][full_path] = []
-                dct['gt']['_targets'][full_path].append(np_targets)
+            reduced_with_grad, reduced = self._reducing_func(outputs=outputs[i],
+                                                             targets=outputs[i]['gt']['_targets'],
+                                                             r_device_metrics=self.r_device_metrics,
+                                                             r_leaf_losses=self.r_leaf_losses,
+                                                             r_per_sample_losses=self.r_per_sample_losses)
+            device_reduced_results.append(reduced_with_grad)
+            for key, value in reduced.items():
+                if key not in ctx_reductions:
+                    ctx_reductions[key] = []
+                value = value.detach().cpu()
+                if len(value.shape) == 0:
+                    value = value.unsqueeze(0)
+                ctx_reductions[key].append(value)
 
-        for callback in self.callbacks:
-            callback.on_dataparallel_gather(dct)
+            if self.store_inference_results:
+                for full_path in self.full_paths:
+                    dct[full_path].append(outputs[i][full_path].detach().cpu().numpy())
+                    precondition_path = f'precondition|{full_path}'
+                    dct[precondition_path].append(outputs[i][precondition_path].detach().cpu().numpy())
+                    np_targets = outputs[i]['gt']['_targets'][full_path].detach().cpu().numpy()
+                    if full_path not in dct['gt']['_targets']:
+                        dct['gt']['_targets'][full_path] = []
+                    dct['gt']['_targets'][full_path].append(np_targets)
+
+        if self.infer_dict_callback is not None:
+            self.infer_dict_callback.on_dataparallel_gather(dct)
         gathered = super().gather(device_reduced_results, output_device)
+        additional_metrics = {key: torch.cat(value, dim=0) for key, value in ctx_reductions.items()}
+
+        for key, value in additional_metrics.items():
+            self.ctx[key] = value
         return gathered
 
+    def reset_device_reducing_tasks(self):
+        self.r_device_metrics = False
+        self.r_leaf_losses = False
+        self.r_per_sample_losses = False
+        self.store_inference_results = len(self.callbacks) > 0
+
 
 class ReplaceGatherCallback(Callback):
 
-    def __init__(self, task_flow, on_gather_callbacks=None):
+    def __init__(self, task_flow, infer_dict_callback=None):
         super().__init__(CallbackOrder.External)
-        if on_gather_callbacks is None:
-            on_gather_callbacks = []
         self.task_flow = task_flow
-        self.on_gather_callbacks = on_gather_callbacks
+        self.infer_dict_callback = infer_dict_callback
 
     def on_stage_start(self, runner: "IRunner"):
         if isinstance(runner.model, DataParallel):
-            runner.model = DeviceReducingDataParallel(runner.model.module, self.task_flow, self.on_gather_callbacks)
+            runner.model = DeviceReducingDataParallel(runner.model.module, self.task_flow, self.infer_dict_callback)
+
+    def on_loader_start(self, runner: "IRunner"):
+        model = runner.model
+        if not isinstance(model, DeviceReducingDataParallel):
+            return
+        for idx, callback in runner.callbacks.items():
+            if isinstance(callback, InterpretationCallback):
+                model.r_per_sample_losses = not should_skip_loader(runner, callback.loaders_to_skip)
+            if isinstance(callback, BatchMetricCallback):
+                model.r_device_metrics = True
+                model.r_leaf_losses = True
 
 
 def reduce_on_device(criterion,
@@ -355,14 +398,52 @@ def reduce_on_device(criterion,
                      leaf_criterions,
                      metrics,
                      outputs,
-                     targets):
+                     targets,
+                     r_device_metrics,
+                     r_leaf_losses,
+                     r_per_sample_losses):
     loss = criterion(outputs, targets)
     any_tensor = any_value(targets)
     n = len(any_tensor)
-    reduced = {
-        '_device|overall|loss': loss,
-        '_device|overall|_n': torch.tensor(n, dtype=any_tensor.dtype, device=any_tensor.device)
+    criterion_n = torch.tensor(n, dtype=any_tensor.dtype, device=any_tensor.device)
+    reduced_with_grad = {
+        f'_device|{criterion.prefix}{criterion.task_name}|loss': loss,
+        f'_device|{criterion.prefix}{criterion.task_name}|_n': criterion_n,
+        f'_device|overall|loss': loss,
+        f'_device|overall|_n': criterion_n
     }
+    reduced = {}
+
+    with torch.no_grad():
+        if r_device_metrics:
+            compute_device_metrics(reduced, any_tensor, metrics, outputs, targets)
+        if r_leaf_losses:
+            compute_leaf_losses(leaf_criterions, outputs, reduced, targets)
+        if r_per_sample_losses:
+            compute_per_sample_losses(reduced, per_sample_criterion, outputs, targets, n)
+
+    return reduced_with_grad, reduced
+
+
+def compute_leaf_losses(leaf_criterions, outputs, reduced, targets):
+    for path, leaf_loss in leaf_criterions.items():
+        reduced[f'_device|{path}|loss'] = leaf_loss(outputs, targets).loss_items
+        reduced[f'_device|{path}|_n'] = outputs[f'precondition|{path}'].sum()
+
+
+def compute_per_sample_losses(reduced, per_sample_criterion, outputs, targets, n):
+    per_sample_losses = per_sample_criterion(outputs, targets)
+    for key, value in per_sample_losses.items():
+        if key.startswith('indices'):
+            value += (n * value.device.index)
+        if len(value.shape) == 0:
+            value = value.unsqueeze(0)
+        reduced[f'_device|{key}|loss_per_sample'] = value
+        if not key.startswith('indices') and key != 'overall':
+            reduced[f'_device|{key}|_n'] = outputs[f'precondition|{key}'].sum()
+
+
+def compute_device_metrics(reduced, any_tensor, metrics, outputs, targets):
     for metric_name, metric in metrics:
         path = f'{metric.prefix}{metric.task_name}'
         full_name = f'{path}|{metric_name}'
@@ -379,16 +460,3 @@ def reduce_on_device(criterion,
                 value = value.unsqueeze(0)
             reduced[f'_device|{full_name}'] = value
         reduced[f'_device|{path}|_n'] = outputs[f'precondition|{metric.prefix}{metric.task_name}'].sum()
-    per_sample_losses = per_sample_criterion(outputs, targets)
-    for key, value in per_sample_losses.items():
-        if key.startswith('indices'):
-            value += (n * value.device.index)
-        if len(value.shape) == 0:
-            value = value.unsqueeze(0)
-        reduced[f'_device|{key}|loss_per_sample'] = value
-        if not key.startswith('indices') and key != 'overall':
-            reduced[f'_device|{key}|_n'] = outputs[f'precondition|{key}'].sum()
-    for path, leaf_loss in leaf_criterions.items():
-        reduced[f'_device|{path}|loss'] = leaf_loss(outputs, targets).loss_items
-        reduced[f'_device|{path}|_n'] = outputs[f'precondition|{path}'].sum()
-    return reduced
\ No newline at end of file
