@@ -1,3 +1,4 @@
+import torch
 
 
 def positive_values(tensor):
@@ -6,3 +7,7 @@ def positive_values(tensor):
     if len(axes) > 0:
         mask = mask.sum(axis=axes) > 0
     return mask
+
+
+def all_correct(tensor):
+    return torch.ones(len(tensor), dtype=bool)
