@@ -1,32 +1,26 @@
 from collections import OrderedDict
+from functools import partial
 from pathlib import Path
 
+import cv2
 import numpy as np
 import pandas as pd
-import cv2
 import torch
-
-from functools import partial
-
-from joblib import delayed
+from torch import nn
 from torch.utils.data import DataLoader
 
 from dnn_cool.converters import TypeGuesser, ValuesConverter, TaskConverter, Converters
-from dnn_cool.decoders import Decoder, BoundedRegressionDecoder
+from dnn_cool.decoders import BoundedRegressionDecoder
 from dnn_cool.project import Project
-from dnn_cool.task_converters import To, ToEfficient
-from dnn_cool.task_flow import BoundedRegressionTask, BinaryClassificationTask, TaskFlow, ClassificationTask, \
+from dnn_cool.task_converters import To
+from dnn_cool.task_flow import BoundedRegressionTask, BinaryClassificationTask, ClassificationTask, \
     MultilabelClassificationTask
 from dnn_cool.utils import torch_split_dataset
-from dnn_cool.value_converters import binary_value_converter, classification_converter, multilabel_converter, \
-    ImageCoordinatesValuesConverter
-from torch import nn
+from dnn_cool.value_converters import binary_value_converter, classification_converter, \
+    ImageCoordinatesValuesConverter, MultiLabelValuesConverter
 
 
-def generate_camera_blocked_image():
-    img = np.zeros((64, 64, 3), dtype=np.uint8)
-
-    # randomly add some noise
+def add_random_noise(img):
     n = 40
     i = np.random.randint(0, 64, size=n)
     j = np.random.randint(0, 64, size=n)
@@ -34,20 +28,18 @@ def generate_camera_blocked_image():
     values = np.random.randint(0, 255, size=n)
     img[i, j, c] = values
 
+
+def generate_camera_blocked_image():
+    img = np.zeros((64, 64, 3), dtype=np.uint8)
+    # randomly add some noise
+    add_random_noise(img)
     return img, {'camera_blocked': True}
 
 
 def generate_door_open_image():
     img = np.ones((64, 64, 3), dtype=np.uint8) * 255
-
     # randomly add some noise
-    n = 40
-    i = np.random.randint(0, 64, size=n)
-    j = np.random.randint(0, 64, size=n)
-    c = np.random.randint(0, 3, size=n)
-    values = np.random.randint(0, 255, size=n)
-    img[i, j, c] = values
-
+    add_random_noise(img)
     return img, {'camera_blocked': False, 'door_open': True, 'person_present': False}
 
 
@@ -62,8 +54,8 @@ def generate_door_closed_image(door_locked):
 
     if door_locked:
         lock_start = x1, int((y1 + y2) / 2)
-        lock_end = lock_start[0] + 30, lock_start[1] + 30
-        img = cv2.rectangle(img, lock_start, lock_end, color=(255, 255, 255), thickness=-1)
+        lock_end = lock_start[0] + 10, lock_start[1] + 10
+        img = cv2.rectangle(img, lock_start, lock_end, color=(0, 0, 255), thickness=-1)
 
     res['door_locked'] = door_locked
     res['camera_blocked'] = False
@@ -175,14 +167,14 @@ def create_df_and_images_tensor(n=int(1e4), cache_file=Path('dnn_cool_synthetic_
         names.append(f'{i}.jpg')
 
     df = pd.DataFrame(rows)
-    df['img'] = names
+    df['syn_img'] = names
     df.loc[:5, 'camera_blocked'] = np.nan
     res = torch.stack(imgs, dim=0).float() / 255., df
     torch.save(res, cache_file)
     return res
 
 
-def synthenic_dataset_preparation(n=int(1e4)):
+def synthetic_dataset_preparation(n=int(1e4)):
     imgs, df = create_df_and_images_tensor(n)
     output_col = ['camera_blocked', 'door_open', 'person_present', 'door_locked',
                   'face_x1', 'face_y1', 'face_w', 'face_h',
@@ -201,7 +193,7 @@ def synthenic_dataset_preparation(n=int(1e4)):
     type_guesser.type_mapping['body_y1'] = 'continuous'
     type_guesser.type_mapping['body_w'] = 'continuous'
     type_guesser.type_mapping['body_h'] = 'continuous'
-    type_guesser.type_mapping['img'] = 'img'
+    type_guesser.type_mapping['syn_img'] = 'img'
     type_guesser.type_mapping['shirt_type'] = 'category'
     type_guesser.type_mapping['facial_characteristics'] = 'multilabel'
 
@@ -210,6 +202,8 @@ def synthenic_dataset_preparation(n=int(1e4)):
     values_converter.type_mapping['binary'] = binary_value_converter
     values_converter.type_mapping['continuous'] = ImageCoordinatesValuesConverter(dim=64)
     values_converter.type_mapping['category'] = classification_converter
+
+    multilabel_converter = MultiLabelValuesConverter()
     values_converter.type_mapping['multilabel'] = multilabel_converter
 
     task_converter = TaskConverter()
@@ -221,17 +215,20 @@ def synthenic_dataset_preparation(n=int(1e4)):
                                                    decoder_supplier=partial(BoundedRegressionDecoder, scale=64))
     n_classes = classification_converter(df['shirt_type']).max().item() + 1
     task_converter.type_mapping['category'] = To(ClassificationTask,
-                                                 module_supplier=partial(nn.Linear, in_features=256, out_features=n_classes))
+                                                 module_supplier=partial(nn.Linear,
+                                                                         in_features=256, out_features=n_classes))
     n_classes = multilabel_converter(df['facial_characteristics']).shape[1]
     task_converter.type_mapping['multilabel'] = To(MultilabelClassificationTask,
-                                                   module_supplier=partial(nn.Linear, in_features=256, out_features=n_classes))
+                                                   module_supplier=partial(nn.Linear,
+                                                                           in_features=256, out_features=n_classes))
 
     converters = Converters()
     converters.task = task_converter
     converters.type = type_guesser
     converters.values = values_converter
 
-    project = Project(df, input_col='img', output_col=output_col, converters=converters, project_dir='./security_project')
+    project = Project(df, input_col='syn_img', output_col=output_col, converters=converters,
+                      project_dir='./security_project')
 
     @project.add_flow
     def face_regression(flow, x, out):
@@ -323,7 +320,7 @@ def synthenic_dataset_preparation(n=int(1e4)):
 
         def forward(self, x):
             res = {}
-            common = self.seq(x['img'])
+            common = self.seq(x['syn_img'])
             res['features'] = self.features_seq(common)
             res['face_localization'] = self.face_localization_seq(common)
             res['body_localization'] = self.body_localization_seq(common)
